/*	$Id: lmin.c,v 3.6 1997/08/13 12:48:19 kas Exp $	*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "proto.h"
#include "ff.h"
#include "pval.h"

/*** Headers for local minimization procedures written in FORTRAN ***/

#ifdef FLONGINT
typedef long int integer; /* Fortran integer */
typedef long int ftnlen;  /* string length in Fortran */
#else
typedef int integer;
typedef int ftnlen; 
#endif

#ifdef NAG
/* 1. Conjugate gradient with limited memory BFGS for updates from
   NAG Fortran Library Mark 16. */

void e04dgf_(integer *n, 
  void (*fungrd)(integer *mode, integer *n, double x[], double *objf, 
                 double grd[], integer *nstate, integer iuser[], double user[]),
  integer *iter, double *objf, double objgrd[], double x[], integer iwork[], 
  double work[], integer iuser[], double user[], integer *ifail);

void e04dkf_(char *string, ftnlen length_1);

static void field(integer *mode, integer *nvars, double vars[], double *objf, 
  double grd[], integer *nstate, integer iuser[], double user[]);

static potential_value *my_pvp;

#else
/* 2. Limited memory BFGS from MINPACK-2 (directory vmlm). */

void dvmlm_(integer *n, double x[], double *objf, double objgrd[], 
  double *frtol, double *fatol, double *fmin, char *task, 
  integer *m, double s[], double y[], double rho[], 
  integer isave[5], double dsave[24],
  double wa1[], double wa2[], ftnlen task_l);

static void field(potential_value *pvp, double *objf, double grd[]);
#endif

#define max(A, B) ((A) > (B) ? (A) : (B))
#define DET(a11,a12,a21,a22) ((a11)*(a22)-(a21)*(a12))

REAL minimize_l(pvp, precision)
potential_value *pvp;
double precision; /* absolute accuracy of the result */
{
	integer dim;

#ifdef NAG
        double *w, *grad, *scratch, f;
        integer *iw, liw, lw, niter, ifail, *iscratch;
        double maxstep;
	char str[50];
#else /* Minpack */
	double *grad, *s, *y, *rho, *wa1, *wa2, f, dsave[24];
	double frtol=1.0e-12, fatol=1.0e-10, fmin=-1.0e30;
	char task[60];
	integer m=5, isave[5];
	ftnlen task_l;
#endif

	if (sizeof(REAL) != sizeof(double)) {
	        perror("lmin.c: REAL must be double");
	        exit(1);
	}
        dim = pvp->ffp->natoms*3; /* number of variables */

#ifdef NAG
        liw = dim+2; /* length of integer workspace */
        lw = 13*dim; /* length of floating-point workspace (double) */

        /* allocate workspace required by e04dgf_ : */
        if ((( grad = (double *) malloc(sizeof(double)*dim)) == NULL) ||
            ((iw = (integer *) malloc(sizeof(int)*liw)) == NULL)      ||
            (( w = (double *) malloc(sizeof(double)*lw)) == NULL)     ||
            ((scratch = (double *) malloc(sizeof(double)*4*dim)) == NULL) ||
            ((iscratch = (integer *) malloc(sizeof(int)*dim)) == NULL) )

                return pvp->value;

        my_pvp=pvp;
        ifail = -1; /* soft return in the case of failure */

        e04dkf_("Nolist", 6);
        e04dkf_("Print Level =  0", 16);
        e04dkf_("Verify Level = -1", 17);
        maxstep = 0.75;
        sprintf(str,"Maximum Step = %5.3f", maxstep);
        e04dkf_(str, 20);
        e04dkf_("Linesearch Tolerance = 0.9", 26);
        sprintf(str,"Optimality Tolerance = %8.3e", precision);
        e04dkf_(str, 31);

        e04dgf_(&dim, field, &niter, &f, grad, pvp->x, iw, w, 
                iscratch, scratch, &ifail);

        pvp->value = f;

        free(grad);
        free(iw), free(w);
        free(iscratch), free(scratch);

#else /* Minpack */
	fatol = precision;

	/* allocate workspace required by dvmlm_ : */
	if ((( grad = (double *) malloc(sizeof(double)*dim)) == NULL) ||
	    ((s = (double *) malloc(sizeof(double)*dim*m)) == NULL)   ||
	    ((y = (double *) malloc(sizeof(double)*dim*m)) == NULL)   ||
	    ((rho = (double *) malloc(sizeof(double)*m)) == NULL)     ||
	    ((wa1 = (double *) malloc(sizeof(double)*dim)) == NULL)   ||
	    ((wa2 = (double *) malloc(sizeof(double)*m)) == NULL) )

		return pvp->value;

	/* Initial function & gradient evaluation: */
	
	field(pvp, &f, grad);
	
	/* Initial call to dvmlm_: */

	strcpy(task, "START"); task_l = 5;
	do {
		/* Minimization step: */
		dvmlm_(&dim, pvp->x, &f, grad, &frtol, &fatol, &fmin,
		       task, &m, s, y, rho,
        	       isave, dsave, wa1, wa2, task_l);

		if (strncmp(task, "FG", 2) == 0) 
			field(pvp, &f, grad);
		else if (strncmp(task, "NEWX", 4) == 0) 
			; /* printf("%f\n", f); */
		else break;

	} while (1);
	
        pvp->value = f;

	free(grad);
	free(s); free(y); free(rho); free(wa1); free(wa2);
#endif

	return compute_potential(pvp);
}

void solve3lin(double a[3][3], double b[], double x[])
/* solve a regular well-conditioned system of three linear equations ax=b */

{
	double det;
	int i,j;

	det = 0.0;
	for(i=0;i<3;i++) det += a[i][0] * DET(a[(i+1)%3][1], a[(i+1)%3][2],
					      a[(i+2)%3][1], a[(i+2)%3][2]);
	for(j=0;j<3;j++) {
		x[j] = 0.0;
		for(i=0;i<3;i++)
		  x[j] += b[i] * DET(a[(i+1)%3][(j+1)%3], a[(i+1)%3][(j+2)%3],
				     a[(i+2)%3][(j+1)%3], a[(i+2)%3][(j+2)%3]);
		x[j] /= det;
	}
}

#ifdef NAG
static void field(integer *mode, integer *nvars, double vars[], double *objf, 
		  double grd[], integer *nstate, integer iscratch[], 
		  double scratch[])
#else /* Minpack */
static void field(potential_value *pvp, double *objf, double grd[])
#endif
{
	int i, dim;
        double a[3][3], b[3], x[3], value;
        int natom;
        double nx, ny, nz;
	static double delta=0.05;
  #ifdef PRINT_DEBUG
        static int fcount=0, gcount=0;
  #endif

#ifdef NAG
        static REAL *xyz0;
#else /* Minpack */
	double *vars;
	potential_value *my_pvp;
#endif

#ifdef NAG
        /* Save the address of the vector of variables in my_pvp: */
        xyz0 = my_pvp->x;
        /* Substitute the vector comming from the optimization procedure
           for the original vector of variables in my_pvp: */
        my_pvp->x = vars;

	dim = *nvars;

        if (vars != my_pvp->x) {
                printf ("Invalid call of field() -- need to save pvp->x");
                exit(1);
        }
        if (*nvars != 3*my_pvp->ffp->natoms) {
                printf ("Weird: In field(): Got array of %d numbers "
                        "(expecting %d from my_pvp->ffp->natoms)\n",
                        dim, 3*my_pvp->ffp->natoms);
                exit(1);
        }

        if (*mode == 0 || *mode == 2) {

#else /* Minpack */

        dim = 3*pvp->ffp->natoms;
	vars = pvp->x;
	my_pvp = pvp;
#endif

		/* Compute potential: */

  	#ifdef PRINT_DEBUG
		printf("field %d: starting\n",++fcount);
  	#endif
		value = *objf = compute_potential(my_pvp);
	#ifdef PRINT_DEBUG
                printf("field %d: %10.5f\n",fcount,*objf);
		for(i=0;i<dim;i++) printf("%7.3f",vars[i]); printf("\n");
	#endif

#ifdef NAG
        }
        if (*mode == 2) { /* compute gradient */
#endif
		/* Compute gradient: */

	#ifdef PRINT_DEBUG
                printf("field grad %d\n",++gcount);
	#endif
                b[0] = value;
                for(i=0;i<dim;i++) {
                        a[0][0] = vars[i]*vars[i]; a[0][1] = vars[i];
                        a[0][2] = a[1][2] = a[2][2] = 1.0;
                        natom = i % my_pvp->ffp->natoms; 
			/* natom = number of the atom that moves */
                        vars[i] -= delta;
                        nx = my_pvp->x[natom]; ny =  my_pvp->y[natom]; nz = my_pvp->z[natom];
                        value = recompute_potential(my_pvp, natom, nx, ny, nz);
                        b[1] = value; a[1][0] = vars[i]*vars[i]; a[1][1] = vars[i];
                        vars[i] += 2.0*delta;
                        nx = my_pvp->x[natom]; ny =  my_pvp->y[natom]; nz = my_pvp->z[natom];
                        value = recompute_potential(my_pvp, natom, nx, ny, nz);
                        b[2] = value; a[2][0] = vars[i]*vars[i]; a[2][1] = vars[i];
                        vars[i] -= delta;
                        solve3lin(a,b,x);

                        grd[i] = 2*x[0]*vars[i]+x[1];
                }
#ifdef NAG
        }
        /* Restore the old vector of variables: */
        my_pvp->x = xyz0;
#endif
}
